// flop.c
//
// Flop is Copyright (c) 1995,2010 by Jason Summers
// This program is distributed under the terms of the GNU General
// Public License, version 3 or higher.

#define WINVER         0x0500
#define _WIN32_WINNT   0x0500
#define _WIN32_WINDOWS 0x0500

#include <windows.h>
#include <tchar.h>
#include <commctrl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "flop.h"
#include "resource.h"
#include <strsafe.h>

#define FL_GET_X_LPARAM(lp)  ((int)(short)LOWORD(lp))
#define FL_GET_Y_LPARAM(lp)  ((int)(short)HIWORD(lp))

#define FL_XOFFS 1
#define FL_YOFFS 1

static INT_PTR CALLBACK DlgProcAbout(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK DlgProcGetText(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam);
static INT_PTR CALLBACK DlgProcGetNumber(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam);

static void Retile(struct Ctx *ctx);
static int fl_GetText(struct Ctx *ctx, const TCHAR *, int);
static void ShowLoc(struct Ctx *ctx, HDC hdc, int i,int j,int force);
static void ShowLocNoDC(struct Ctx *ctx, int i,int j,int force);

struct Game game;      // Current game position
struct Game newgame;   // Game position at the start of the current level

static const int g_direc[4][2]= { {0,-1},{1,0},{0,1},{-1,0} };


static int fl_ConstrainInt(int x, int xmin, int xmax)
{
	if(x<xmin) return xmin;
	if(x>xmax) return xmax;
	return x;
}

static int fl_random(int __num)
{
	return rand()%__num;
}

void Error(struct Ctx *ctx, const TCHAR *m)
{
	MessageBox(ctx?ctx->hwndMain:NULL,m,_T("Error"),MB_OK|MB_ICONEXCLAMATION);
}

// copy n moves     note: can always copy to lower address
static void CopyMoves(unsigned char *f,unsigned char *t, int n)
{   int i;
	 for(i=0; i<n; i++) t[i]=f[i];
//    t->movesvalid=f->movesvalid;
//    t->moves=0;
}

// x==0: moves & rings only           x==1: everything
static void UpdateStat(struct Ctx *ctx, int x)
{
	if(ctx->hwndMoves) InvalidateRect(ctx->hwndMoves,NULL,TRUE);
	if(x && ctx->hwndTitle) InvalidateRect(ctx->hwndTitle,NULL,TRUE);
}

void UpdateAll(struct Ctx *ctx)
{
	UpdateStat(ctx,1);
	Retile(ctx);
}


/****************************************************************************/

// Initializes a Game structure.
void fl_InitGame(struct Game *g)
{
	int i,j;

	ZeroMemory(g,sizeof(struct Game));

	for(i=0;i<TX+2;i++)         // this ensures the border is all tiles
		for(j=0;j<TY+2;j++)
			g->field[i][j]=I_TILE;
	g->field[1][1]=0;

	g->moves=0;
    g->movesvalid=0;
	g->moveptr=0;
	g->rings=0;   // number of rings collected

	g->pos.x=1; g->pos.y=1;
	g->needed=0;
    StringCchCopy(g->levelname,FL_MAXNAMELEN+1,_T("No level loaded"));
	StringCchCopy(g->message,FL_MAXMESSAGELEN+1,_T("Good!"));
	g->bestscore= -1;
	StringCchCopy(g->bestname,FL_MAXNAMELEN+1,_T("(none)"));

	g->tileselected=0;
	g->xs=0;	// number of uncovered X's
	g->levelcomplete=0;
}

static void CheckComplete(struct Ctx *ctx, int needtoredraw)	// check for completed game
{
	if(game.xs<1 && game.rings>=game.needed &&
			(game.field[game.pos.x][game.pos.y]&I_DOOR) )
	{
		game.levelcomplete=1;

		if(needtoredraw) {
			Retile(ctx);
		}
		else {
			ShowLocNoDC(ctx,game.pos.x,game.pos.y,0);
		}

		MessageBox(ctx->hwndMain,game.message,_T("Level complete!"),MB_OK|MB_ICONASTERISK);

		if(newgame.bestscore<0 || game.moves<newgame.bestscore) { // new low score
			StringCchCopy(ctx->textbuf,256,ctx->playername);
			fl_GetText(ctx,_T("You have the best score for this level; please enter your name:"),FL_MAXNAMELEN);
			StringCchCopy(newgame.bestname,FL_MAXNAMELEN+1,ctx->textbuf);
			StringCchCopy(game.bestname,FL_MAXNAMELEN+1,ctx->textbuf);
			StringCchCopy(ctx->playername,FL_MAXNAMELEN+1,ctx->textbuf);
			newgame.bestscore=game.moves;
			game.bestscore=game.moves;
			SaveLevel(ctx,&newgame,ctx->level,1);
			UpdateStat(ctx,1);
		}
	}
}

void fl_NewGame(struct Ctx *ctx)  // copies newgame to game and inits things for a level restart
{
	int i,j;
	game=newgame;
	game.xs=0;
	for(i=1;i<=TX;i++)		// count uncovered X's
		for(j=1;j<=TY;j++)
			if(game.field[i][j]==I_X) game.xs++;
	game.levelcomplete=0;
	game.tileselected=0;
}

// returns 1 if user requests cancel
// returns 0 if operation should continue
static int WarnChanged(struct Ctx *ctx)
{
	int x;

	if(!ctx->dirty || !ctx->editing) return 0;
	x=MessageBox(ctx->hwndMain,_T("Lose changes made to this level?"),_T("Warning"),
		MB_OKCANCEL|MB_ICONEXCLAMATION);
	if(x==IDOK) ctx->dirty=0;
	return (x==IDCANCEL);
}


// Sets ctx->movewindowtextlen
static void CalcMovesWindowWidth(struct Ctx *ctx, HWND hWnd)
{
	HDC hDC;
	static const TCHAR texttemplate[] = _T(" Rings: 00/00   Moves: 000 (000) ");
	SIZE size;

	if(ctx->movewindowtextlen>0) return; // already done

	hDC=GetDC(hWnd);
	SelectObject(hDC,ctx->hStatFont);
	GetTextExtentPoint32(hDC,texttemplate,lstrlen(texttemplate),&size);
	ctx->movewindowtextlen = size.cx;
	ReleaseDC(hWnd,hDC);
}

// sets up the variables Lx, Ly (before drawing some tiles)
static void CalcTileSizes(struct Ctx *ctx)
{
	RECT r;

	if(ctx->hwndGame) {
		GetClientRect(ctx->hwndGame,&r);
	}
	else {
		r.left=r.top=0;
		r.right=r.bottom=TX*10;
	}

	ctx->Lx=(r.right-3)/TX;
	ctx->Ly=(r.bottom-3)/TY;
	if(ctx->Lx<5) ctx->Lx=5;
	if(ctx->Ly<5) ctx->Ly=5;

	ctx->y_handle_offs= (ctx->Ly+50)/30;
	ctx->x_handle_offs= (ctx->Lx+50)/30;
}

static int Flip(int what,int dir) // if direction is even, flip to & bottom, etc.
{
	int t1,t2;

	if(dir&1) {    // direction is odd; flip left & right
		t1= (what&I_HE)?I_HW:0;
		t2= (what&I_HW)?I_HE:0;
		what&= ~(I_HE|I_HW);
		what|= (t1|t2);
	}
	else {
		t1= (what&I_HN)?I_HS:0;
		t2= (what&I_HS)?I_HN:0;
		what&= ~(I_HN|I_HS);
		what|= (t1|t2);
	}
	return what & (I_TILE|I_ALLH);
}

//  Invert the colors of a square.  hMainDC must be valid
static void InvertLoc(struct Ctx *ctx, HDC hdc, int x,int y)
{
	RECT r;

	SetRect(&r,
		FL_XOFFS+(x-1)*ctx->Lx+3,
		FL_YOFFS+(y-1)*ctx->Ly+3,
		FL_XOFFS+x*ctx->Lx-1,
		FL_YOFFS+y*ctx->Ly-1);
	InvertRect(hdc,&r);
}

static void DrawItems(struct Ctx *ctx, HDC hdc,int what,int x,int y,int dooropen)
{
	int x1,x2,y1,y2,xs,ys;
	int xoffs=FL_XOFFS;
	int yoffs=FL_YOFFS;

	x1= xoffs+ (x-1)*ctx->Lx+3;    x2= xoffs+     x*ctx->Lx-1;
	y1= yoffs+ (y-1)*ctx->Ly+3;    y2= yoffs+     y*ctx->Ly-1;
	xs=x2-x1;          ys=y2-y1;

	SelectObject(hdc,GetStockObject(NULL_PEN));   // clear the area
	SelectObject(hdc,ctx->hClearBrush);
	Rectangle(hdc,x1,y1,x2+1,y2+1);
	//Rectangle(hdc,x1,y1,x2-3,y2-3);

	if(what&I_TILE) {
		SelectObject(hdc,ctx->hTilePen);   // draw the tile
		SelectObject(hdc,ctx->hTileBrush);
		Rectangle(hdc,x1,y1,x2,y2);

		SelectObject(hdc,ctx->hHandlePen);
		if(what&I_HN) { // north handle
			MoveToEx(hdc,x1+   xs/4   ,y1+ctx->y_handle_offs,NULL); // y_h... usu. = 2
			LineTo(hdc,x1+(3*xs)/4+1,y1+ctx->y_handle_offs);
		}
		if(what&I_HE) { // east handle
			MoveToEx(hdc,x2-1-ctx->x_handle_offs,y1+   ys/4,NULL);
			LineTo(hdc,x2-1-ctx->x_handle_offs,y1+(3*ys)/4+1);
		}
		if(what&I_HS) { // south handle
			MoveToEx(hdc,x1+   xs/4   ,y2-1-ctx->y_handle_offs,NULL);
			LineTo(hdc,x1+(3*xs)/4+1,y2-1-ctx->y_handle_offs);
		}
		if(what&I_HW) { // west handle
			MoveToEx(hdc,x1+ctx->x_handle_offs,y1+   ys/4,NULL);
			LineTo(hdc,x1+ctx->x_handle_offs,y1+(3*ys)/4+1);
		}
	}
	if(what&I_RING) {
		SelectObject(hdc,ctx->hRingPen);
		SelectObject(hdc,GetStockObject(NULL_BRUSH));
		Ellipse(hdc,x1+(3*xs)/8,y1+(3*ys)/8,
					x1+(5*xs)/8,y1+(5*ys)/8);
	}
	if(what&I_X) {
		SelectObject(hdc,ctx->hXPen);
		MoveToEx(hdc,x1+3,y1+3,NULL);
		LineTo(hdc,x2-3,y2-3);
		MoveToEx(hdc,x1+3,y2-3,NULL);
		LineTo(hdc,x2-3,y1+3);
	}
	if(what&I_DOOR) {
		SelectObject(hdc,ctx->hDoorPen);
		if(dooropen) SelectObject(hdc,GetStockObject(BLACK_BRUSH));
		else         SelectObject(hdc,ctx->hDoorBrush);
		Rectangle(hdc,x1+(3*xs)/8,y1+(2*ys)/8,
					  x1+(5*xs)/8,y1+(6*ys)/8);
	}
	if(what&I_PLAYER) {
		SelectObject(hdc,GetStockObject(NULL_PEN));
		SelectObject(hdc,ctx->hPlayerBrush);
		Ellipse(hdc,x1+(1*xs)/8,y1+(1*ys)/8,
					x1+(7*xs)/8+1,y1+(7*ys)/8+1);
	}
}

// figures out what should be shown at (x,y) and calls DrawItems to
// display it.  (Lx, Ly must be set up)
// draws only if there has been a change, unless force==1
static void ShowLoc(struct Ctx *ctx, HDC hdc, int i,int j, int force)
{
	int toshow, opendoor=0;

	toshow=game.field[i][j];
	if(ctx->editing) {
		if(i==game.pos.x && j==game.pos.y)
			toshow |= I_PLAYER;
		if(i==ctx->editpos.x && j==ctx->editpos.y)
			toshow |= I_INVERTED;
	}
	else {   // not editing
		opendoor=(game.rings>=game.needed && game.xs<1);
		if(i==game.pos.x && j==game.pos.y) {
			if(game.levelcomplete)
				opendoor=0;
			else
				toshow |= I_PLAYER;
		}
		if(opendoor && (game.field[i][j]&I_DOOR))
			toshow |= I_OPENDOOR;
		if(game.tileselected && i==game.seltile.x && j==game.seltile.y)
			toshow |= I_INVERTED;
	}

	if( (toshow!=ctx->vis[i][j]) || force)
	{
		DrawItems(ctx,hdc,toshow&511,i,j,opendoor);
		if(toshow&I_INVERTED) InvertLoc(ctx,hdc,i,j);
		ctx->vis[i][j]=toshow;
	}
}

// Inefficient but easier-to-call version of ShowLoc().
static void ShowLocNoDC(struct Ctx *ctx, int i,int j, int force)
{
	HDC hdc;

	hdc=GetDC(ctx->hwndGame);
	ShowLoc(ctx,hdc,i,j,force);
	ReleaseDC(ctx->hwndGame,hdc);
}

static void DrawDoors(struct Ctx *ctx)
{
	int i,j;
	HDC hdc;

	hdc=GetDC(ctx->hwndGame);
	for(i=1;i<=TX;i++)
		for(j=1;j<=TY;j++)
			if(game.field[i][j]&I_DOOR) ShowLoc(ctx,hdc,i,j,0);
	ReleaseDC(ctx->hwndGame,hdc);
}

static void Retile(struct Ctx *ctx)    // re-draw every tile
{
	int i,j;
	HDC hdc;

	hdc=GetDC(ctx->hwndGame);
	for(i=1;i<=TX;i++)
		for(j=1;j<=TY;j++) ShowLoc(ctx,hdc,i,j,0);
	ReleaseDC(ctx->hwndGame,hdc);
}

static void SaveMove(struct Ctx *ctx, unsigned char m, unsigned char * xxx)     // m is the move
{
	if(game.movesvalid<=game.moveptr || ctx->move[game.moveptr]!=m) {
		ctx->move[game.moveptr++]=m;
		game.movesvalid=game.moveptr;
	}
	else game.moveptr++;
}


/* Play n moves (stored in movebuf)
 *
 *
 * v: 0==draw only after all moves    1==draw every move
 * returns 1 on success, 0 on error
 */
static int Play(struct Ctx *ctx, struct Game *g ,unsigned char *movebuf,int n,int v)
{   int i,item,res=1;
	//int ptr;     // pointer into movebuf
	unsigned char arrow;
	POINT newpos;
	HDC hdc;

//	if(n<2) v=1;   // if only one move, we always draw after every move
	if(n>10 && v) SetCursor(LoadCursor(NULL,IDC_WAIT));

	for(i=0;i<n;i++) {
		if(g->levelcomplete) {
//            Error(ctx,_T("This level is complete"));
			res=0;
			break;
		}
		if(g->moveptr>=MOVEMAX) {
			Error(ctx,_T("Out of move memory"));
			res=0;
			break;
		}

		arrow=movebuf[i];
		if(arrow>3) {
			Error(ctx,_T("Internal: Bad move data"));
			g->movesvalid=0;      // should be unneeded
			res=0;
			break;
		}

		if(g->tileselected) {   // trying to move a tile
			newpos.x=g->seltile.x+g_direc[arrow][0];
			newpos.y=g->seltile.y+g_direc[arrow][1];
			if(v) {
				ShowLocNoDC(ctx,g->seltile.x,g->seltile.y,0);
			}
			item=g->field[newpos.x][newpos.y];
			if(!(item&I_TILE) &&
				  (g->pos.x!=newpos.x ||
				   g->pos.y!=newpos.y)) {  // we can move tile

				g->field[newpos.x][newpos.y] |=
					Flip(g->field[g->seltile.x][g->seltile.y],arrow);
				g->field[g->seltile.x][g->seltile.y] &= (~(I_TILE|I_ALLH));
				g->tileselected=0;

				if(v) {
					hdc=GetDC(ctx->hwndGame);
					ShowLoc(ctx,hdc,g->seltile.x,g->seltile.y,0);
					ShowLoc(ctx,hdc,newpos.x,newpos.y,0);
					ReleaseDC(ctx->hwndGame,hdc);
				}

				if(g->field[g->seltile.x][g->seltile.y]&I_X) {
					g->xs++;
					if(g->xs==1 && g->rings>=g->needed)
						if(v) DrawDoors(ctx);
				}
				if(g->field[newpos.x][newpos.y]&I_X) {
					g->xs--;
					if(g->xs==0 && g->rings>=g->needed)
						if(v) DrawDoors(ctx);
				}

				// save this move (of a tile)
				SaveMove(ctx,arrow,movebuf);    // SaveMove(arrow);
				UpdateStat(ctx,0);
				CheckComplete(ctx,!v);
			}
			else {
				g->moveptr--; // last move didn't count
				g->tileselected=0;
				UpdateStat(ctx,0);

				if(v) {
					ShowLocNoDC(ctx,g->seltile.x,g->seltile.y,0);
				}
			}
			g->tileselected=0;
		}
		else {  /* tile not selected: trying to move the player */
			item=g->field[g->pos.x+g_direc[arrow][0]]
						   [g->pos.y+g_direc[arrow][1]];
			if(item&I_TILE) {     /* moving into a tile */
				if(   ((arrow==0)&&(item&I_HS)) ||
					  ((arrow==1)&&(item&I_HW)) ||
					  ((arrow==2)&&(item&I_HN)) ||
					  ((arrow==3)&&(item&I_HE))    ) {
					g->tileselected=1;
					g->seltile.x=g->pos.x+g_direc[arrow][0];
					g->seltile.y=g->pos.y+g_direc[arrow][1];
					// highlight this tile
					if(v) {
						ShowLocNoDC(ctx,g->seltile.x,g->seltile.y,0);
					}
					SaveMove(ctx,arrow,movebuf);
					UpdateStat(ctx,0);
				}
				// else player is blocked by tile: do nothing
//				else{      // if stored move is invalid, kill it
//					if(g->moveptr<g->movesvalid){
//						CopyMoves(&movebuf[g->moveptr+1],
//								  &movebuf[g->moveptr],
//								  g->movesvalid-g->moveptr-1);
//						g->movesvalid--;
//					}
//				}
			}
			else {   // move player
				SaveMove(ctx,arrow,movebuf);
				g->pos.x+=g_direc[arrow][0];
				g->pos.y+=g_direc[arrow][1];

				item=g->field[g->pos.x][g->pos.y];
				if(item&I_RING) {
					g->rings++;
					UpdateStat(ctx,0);
					item &= (~I_RING);  // remove the ring
					g->field[g->pos.x][g->pos.y]=item;
					if(g->rings==g->needed && g->xs<=0)
						if(v) DrawDoors(ctx);
				}
				if(v) {
					hdc=GetDC(ctx->hwndGame);
					ShowLoc(ctx,hdc,g->pos.x-g_direc[arrow][0],  // the old position
							g->pos.y-g_direc[arrow][1],0);
					ShowLoc(ctx,hdc,g->pos.x,g->pos.y,0);      // the new
					ReleaseDC(ctx->hwndGame,hdc);
				}
				g->moves++;
				UpdateStat(ctx,0);
				CheckComplete(ctx,!v);
			}
		}
	}
	if((!v) && (!g->levelcomplete)) Retile(ctx);
	if(n>10 && v) SetCursor(LoadCursor(NULL,IDC_ARROW));
	return res;
}

// play the next n saved moves
static void Redo(struct Ctx *ctx, int n)
{
	if(game.moveptr+n > game.movesvalid) {
//		toomany=1;
		n=game.movesvalid-game.moveptr;
	}
	if(n>0) {              // not all moves may be valid, so we must save them
		CopyMoves(&ctx->move[game.moveptr],ctx->move2,n);     // somewhere else and play from there

//		savemoveptr=game.moveptr+n;

		if(n==1) Play(ctx,&game,ctx->move2,n,1);
		else     Play(ctx,&game,ctx->move2,n,!ctx->fastreplay);

		// kluge: if some moves were "bad", we move the good moves down
//        if(game.moveptr<savemoveptr && game.movesvalid>savemoveptr){
//            CopyMoves(&move[savemoveptr],&move[game.moveptr],
//							game.movesvalid-savemoveptr);
//			game.movesvalid-=(savemoveptr-game.moveptr);
//		}

	}
//	if(toomany)
//		Error(ctx,_T("No more moves to replay"));
}

// undo to move n
void Undo(struct Ctx *ctx, int n)
{
	int x,saveval;

	x=n;
	if(x<0) x=0;
	if(x>game.moveptr) {
		Error(ctx,_T("Internal: Can't undo forward"));
		return;
	}
	if(x<game.moveptr) {
		game.tileselected=0;
		saveval=game.movesvalid;
		fl_NewGame(ctx);
		CopyMoves(ctx->move,ctx->move2,n);
		game.movesvalid=saveval;
		//if(x>0)
		Play(ctx,&game,ctx->move2,n,0);
		UpdateStat(ctx,0);
	}
//	if(n<0)
//		Error(ctx,_T("No more moves to undo"));

}

struct fl_gettext_struct {
	int linelengthmax;
	const TCHAR *prompt;
	int success;
	TCHAR textbuf[256];
};

struct fl_getnumber_struct {
	const TCHAR *prompt;
	int startval;
	int minval,maxval;
	int success;
	int return_val;
};

// Asks the user for a string.
// Initial value of the string is in ctx->textbuf.
// Returns text in ctx->textbuf.
// Returns 0 if canceled, 1 of OK.
// maxlength does NOT include the terminating NUL. (max maxlength=255)
static int fl_GetText(struct Ctx *ctx, const TCHAR *prompt,int maxlength)
{
	struct fl_gettext_struct gts;

	gts.prompt = prompt;
	gts.linelengthmax=maxlength;
	gts.success=0;
	StringCchCopy(gts.textbuf,maxlength+1,ctx->textbuf); // FIXME

	DialogBoxParam(ctx->hInst,_T("TextBox"),ctx->hwndMain,DlgProcGetText,(LPARAM)&gts);

	StringCchCopy(ctx->textbuf,maxlength+1,gts.textbuf); // FIXME
	return gts.success;
}

static INT_PTR CALLBACK DlgProcGetText(HWND hDlg,UINT msg,WPARAM wParam,LPARAM lParam)
{
	struct fl_gettext_struct *gts;
	WORD id;

	if(msg==WM_INITDIALOG) {
		SetWindowLongPtr(hDlg,DWLP_USER,lParam);
		gts=(struct fl_gettext_struct*)lParam;
	}
	else {
		gts=(struct fl_gettext_struct*)GetWindowLongPtr(hDlg,DWLP_USER);
	}
	if(!gts) return FALSE;

	switch(msg) {
	case WM_INITDIALOG:
		SendDlgItemMessage(hDlg,IDC_TEXTLINE,EM_LIMITTEXT,gts->linelengthmax,0);
		SetDlgItemText(hDlg, IDC_TEXTLINE, gts->textbuf);
		SetDlgItemText(hDlg, IDC_INPUTPROMPT, gts->prompt);
		return TRUE;

	case WM_COMMAND:
		id=LOWORD(wParam);
		if(id==IDOK) {
			GetDlgItemText(hDlg, IDC_TEXTLINE, gts->textbuf, gts->linelengthmax+1);
			gts->success=1;
		}
		if(id==IDCANCEL) {
			gts->textbuf[0]='\0';
			gts->success=0;
		}
		if(id==IDOK || id==IDCANCEL){
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

// Returns -1 if canceled, the selected number otherwise.
static int fl_GetNumber(struct Ctx *ctx, const TCHAR *prompt,int minnum,int maxnum,int def)
{
	struct fl_getnumber_struct gns;

	gns.prompt=prompt;
	gns.startval=def;
	gns.minval=minnum;
	gns.maxval=maxnum;
	gns.return_val=0;
	gns.success=0;

	DialogBoxParam(ctx->hInst,_T("InputNumberBox"),ctx->hwndMain,DlgProcGetNumber,(LPARAM)&gns);

	if(gns.success) return gns.return_val;
	return -1;
}

static void OnNumINITDIALOG(struct fl_getnumber_struct *gns, HWND hDlg)
{
	SetDlgItemText(hDlg, IDC_INPUTPROMPT, gns->prompt);
	SendDlgItemMessage(hDlg,IDC_TEXTLINE, EM_LIMITTEXT,4,0);
	SendDlgItemMessage(hDlg,IDC_NUMBER_UPDOWN,UDM_SETRANGE,0,(LPARAM)MAKELONG(gns->maxval,gns->minval));
	SendDlgItemMessage(hDlg,IDC_NUMBER_UPDOWN,UDM_SETBUDDY,(WPARAM)GetDlgItem(hDlg,IDC_TEXTLINE),0);
	SendDlgItemMessage(hDlg,IDC_NUMBER_UPDOWN,UDM_SETPOS,0,(LPARAM)MAKELONG(gns->startval,0));
}

static INT_PTR CALLBACK DlgProcGetNumber(HWND hDlg,UINT msg,WPARAM wParam,LPARAM lParam)
{
	struct fl_getnumber_struct *gns;
	WORD id;

	if(msg==WM_INITDIALOG) {
		SetWindowLongPtr(hDlg,DWLP_USER,lParam);
		gns=(struct fl_getnumber_struct*)lParam;
	}
	else {
		gns=(struct fl_getnumber_struct*)GetWindowLongPtr(hDlg,DWLP_USER);
	}
	if(!gns) return FALSE;

	switch(msg) {
	case WM_INITDIALOG:
		OnNumINITDIALOG(gns,hDlg);
		return TRUE;

	case WM_COMMAND:
		id=LOWORD(wParam);
		switch(id) {
		case IDOK:
			gns->return_val=(int)GetDlgItemInt(hDlg, IDC_TEXTLINE, NULL, 0);
			gns->success=1;
			EndDialog(hDlg, TRUE);
			return TRUE;
		case IDCANCEL:
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

static void ButtonClick_Editing(struct Ctx *ctx, HWND hWnd, int x, int y)
{
	HDC hdc;
	int sx,sy;

	hdc=GetDC(hWnd);
	sx=ctx->editpos.x;  sy=ctx->editpos.y;
	ctx->editpos.x=x;   ctx->editpos.y=y;
	ShowLoc(ctx,hdc,sx,sy,0);
	ShowLoc(ctx,hdc,ctx->editpos.x,ctx->editpos.y,0);
	ReleaseDC(hWnd,hdc);
}

static void OnButtonClick(struct Ctx *ctx, HWND hWnd, int xp, int yp)
{
	int x,y;
	unsigned char m,dir; // temporary storage for moves

	// Find the tile that was clicked on.
	x= 1+(xp-2)/ctx->Lx;  x=fl_ConstrainInt(x,1,TX);
	y= 1+(yp-2)/ctx->Ly;
	//if(y>TY) return;   // clicked on status bar
	y=fl_ConstrainInt(y,1,TY);

	if(ctx->editing) {
		ButtonClick_Editing(ctx,hWnd,x,y);
		return;
	}

	if(game.tileselected) {
		// If a tile is selected for moving, adjust the position so that
		// it's relative to the tile instead of the player.
		x+=(game.pos.x-game.seltile.x);
		y+=(game.pos.y-game.seltile.y);
	}
	m=255;
	if(x==game.pos.x) {    // moving directly up or down
		if(y==game.pos.y-1) { m=0; Play(ctx,&game,&m,1,1); }
		if(y==game.pos.y+1) { m=2; Play(ctx,&game,&m,1,1); }
	}
	if(y==game.pos.y) {
		if(x==game.pos.x-1) { m=3; Play(ctx,&game,&m,1,1); }
		if(x==game.pos.x+1) { m=1; Play(ctx,&game,&m,1,1); }
	}
	if(m==255) {
		if(game.tileselected) {    // unselect tile
			game.tileselected=0;
			game.moveptr--;
			UpdateStat(ctx,0);
			ShowLocNoDC(ctx,game.seltile.x,game.seltile.y,0);
		}
		else {            // try to move to the given location
			while(1) {
				dir=255;

				if     (y<game.pos.y && !(game.field[game.pos.x]   // favor vertical moves
													  [game.pos.y-1]&I_TILE)) dir=0;
				else if(y>game.pos.y && !(game.field[game.pos.x]
													  [game.pos.y+1]&I_TILE)) dir=2;
				else if(x>game.pos.x && !(game.field[game.pos.x+1]
													  [game.pos.y]  &I_TILE)) dir=1;
				else if(x<game.pos.x && !(game.field[game.pos.x-1]
													  [game.pos.y]  &I_TILE)) dir=3;

				if(dir>3) break;         // all moves are blocked
				if(!Play(ctx,&game,&dir,1,1)) break;
			}
		}
	}
}

/****************************************************************************/
static void OnRESETTHIS(struct Ctx *ctx)
{
	if(MessageBox(ctx->hwndMain,
		 _T("Are you sure you want to clear the best score for this level?"),
		 _T("Warning"),MB_OKCANCEL|MB_DEFBUTTON2|MB_ICONEXCLAMATION)
		==IDOK)
	{
		newgame.bestscore= -1;
		game.bestscore= -1;
		SaveLevel(ctx,&newgame,ctx->level,1);
		UpdateStat(ctx,1);
	}
}

static void OnEXPORT(struct Ctx *ctx)
{
	if(Export(ctx,&newgame)) {
		MessageBox(ctx->hwndMain,_T("Level exported"),_T(""),MB_OK|MB_ICONINFORMATION);
		ctx->dirty=0;
	}
}

static void AboutBox(struct Ctx *ctx, HWND hWnd)
{
	DialogBox(ctx->hInst,_T("AboutBox"),hWnd,DlgProcAbout);
}

static void LoadAndStartLevel(struct Ctx *ctx, int levelnum)
{
	if(LoadLevel(ctx,&newgame,levelnum)) {
		ctx->level=levelnum;
		fl_NewGame(ctx);
		UpdateAll(ctx);
	}
}

static void OnGOTOLEVEL(struct Ctx *ctx)
{
	int x;
	if(WarnChanged(ctx)) return;
	if((x=fl_GetNumber(ctx,_T("Go to level:"),1,100,ctx->level))>=0) {
		if(x>=1 && x<=100) {
			LoadAndStartLevel(ctx,x);
		}
	}
}

static void OnSAVELEVEL(struct Ctx *ctx)
{
	int i,x;

	if(!ctx->editing) return;

	i= ctx->editall? 1 : 21;
	if((x=fl_GetNumber(ctx,_T("Save as level:"),i,100,ctx->level))>=0) {

		if(x<1 || x>100) { Error(ctx,_T("No such level")); return; }    // won't happen
		if(SaveLevel(ctx,&game,x,0)) {
			MessageBox(ctx->hwndMain,_T("Level saved"),_T(""),MB_OK|MB_ICONINFORMATION);
			ctx->dirty=0;
		}
		ctx->level=x;
		UpdateStat(ctx,1);
	}
}

static void OnLOADMOVES(struct Ctx *ctx)
{
	int x;

	if(ctx->editing) return;

	x=LoadMoves(ctx,ctx->move2,MOVEMAX);
	if(x>0) {
		StringCchPrintf(ctx->textbuf,256,_T("%d keystroke%s loaded"),x,x==1?_T(""):_T("s"));
		MessageBox(ctx->hwndMain,ctx->textbuf,_T(""),MB_OK|MB_ICONINFORMATION);
		Play(ctx,&game,ctx->move2,x,!ctx->fastreplay);
	}
}

static void OnSAVEMOVESFROM(struct Ctx *ctx)
{
	int x;

	if(ctx->editing) return;
	StringCchCopy(ctx->textbuf,256,_T("0"));
	if((x=fl_GetNumber(ctx,_T("Save moves from keystroke number:"),
			0,game.moveptr-1,0))<0)
	{
		return;
	}
	x=fl_ConstrainInt(x,0,game.moveptr);
	SaveMoves(ctx,&ctx->move[x],game.moveptr-x);
}

static void OnGOTOMOVE(struct Ctx *ctx)
{
	int x;
	if(ctx->editing) return;
	if((x=fl_GetNumber(ctx,_T("Go to keystroke number:"),
			   0,game.movesvalid,game.moveptr))>=0)
	{
		if(x<game.moveptr) Undo(ctx,x);
		else if(x>game.moveptr) Redo(ctx,x-game.moveptr);
	}
}

static void OnFASTREPLAY(struct Ctx *ctx)
{
	ctx->fastreplay = !ctx->fastreplay;
	CheckMenuItem(GetMenu(ctx->hwndMain),IDM_FASTREPLAY,
			ctx->fastreplay?MF_CHECKED:MF_UNCHECKED);
}

static void OnEDITNAME(struct Ctx *ctx)
{
	if(!ctx->editing) return;
	StringCchCopy(ctx->textbuf,FL_MAXNAMELEN+1,game.levelname);
	if(fl_GetText(ctx,_T("Name of this level:"),FL_MAXNAMELEN)) {
		StringCchCopy(game.levelname,FL_MAXNAMELEN+1,ctx->textbuf);
		UpdateStat(ctx,1);
		ctx->dirty=1;
	}
}

static void OnEDITMESSAGE(struct Ctx *ctx)
{
	if(!ctx->editing) return;
	StringCchCopy(ctx->textbuf,FL_MAXMESSAGELEN+1,game.message);
	if(fl_GetText(ctx,_T("Message upon completing level:"),FL_MAXMESSAGELEN)) {
		StringCchCopy(game.message,FL_MAXMESSAGELEN+1,ctx->textbuf);
		ctx->dirty=1;
	}
}

static void OnEDITRINGS(struct Ctx *ctx)
{
	int x;
	if(!ctx->editing) return;
	if((x=fl_GetNumber(ctx,_T("Number of rings needed:"),
					  0,144,game.needed))>=0)
	{
		game.needed=x;
		UpdateStat(ctx,1);
		ctx->dirty=1;
	}
}

static void OnEDITCLEAR(struct Ctx *ctx)
{
	int i,j;
	if(!ctx->editing) return;
	if(WarnChanged(ctx)) return;
	for(i=1;i<=TX;i++)
		for(j=1;j<=TY;j++)
			game.field[i][j]=0;
	Retile(ctx);
}

static void OnEDITRANDOM(struct Ctx *ctx)
{
	int i,j;
	int x,y;
	int x1,x2;

	if(!ctx->editing) return;

	if((x1=fl_GetNumber(ctx,_T("Percent chance of a tile (0-100):"),0,100,ctx->random_tile_percent))>=0)
		 ctx->random_tile_percent= x1;
	ctx->random_tile_percent=fl_ConstrainInt(ctx->random_tile_percent,-1,100);

	if((x2=fl_GetNumber(ctx,_T("Percent chance of a handle (0-100):"),0,100,ctx->random_handle_percent))>=0)
		 ctx->random_handle_percent= x2;
	ctx->random_handle_percent=fl_ConstrainInt(ctx->random_handle_percent,-1,100);
	if(x1<0 && x2<0) return; // both were canceled

	for(i=1;i<=TX;i++)
		for(j=1;j<=TX;j++) {
			x=fl_random(100);
			if(x1>=0) {  // don't change if this option was canceled
				if(ctx->random_tile_percent>x) game.field[i][j] |= I_TILE;
				else game.field[i][j] &= (~(I_TILE|I_ALLH));
			}
			if((game.field[i][j]&I_TILE)&&(x2>=0)) {
				y=0;
				x=fl_random(100); if(ctx->random_handle_percent>x) y |= I_HN;
				x=fl_random(100); if(ctx->random_handle_percent>x) y |= I_HS;
				x=fl_random(100); if(ctx->random_handle_percent>x) y |= I_HE;
				x=fl_random(100); if(ctx->random_handle_percent>x) y |= I_HW;
				game.field[i][j] &= (~I_ALLH);
				game.field[i][j] |= y;
			}
		}
	game.field[game.pos.x][game.pos.y]=0;
	Retile(ctx);
	ctx->dirty=1;
}

static void OnEditShift(struct Ctx *ctx, int dir)
{
	int i, j;
	int savefield[TX+2][TY+2];

	if(!ctx->editing) return;

	//x=(int)id-IDM_EDITUP;  // the direction
	for(i=1;i<=TX;i++)
		for(j=1;j<=TY;j++)
			savefield[i][j]=
			   game.field[(i-g_direc[dir][0]-1+TX)%TX+1]
						 [(j-g_direc[dir][1]-1+TY)%TY+1];
	for(i=1;i<=TX;i++)
		for(j=1;j<=TY;j++)
			game.field[i][j]=savefield[i][j];
	game.pos.x=(game.pos.x+g_direc[dir][0]-1+TX)%TX+1;
	game.pos.y=(game.pos.y+g_direc[dir][1]-1+TY)%TY+1;
	Retile(ctx);
	ctx->dirty=1;
}

static void OnEditTile(struct Ctx *ctx, WORD id)
{
	int x;
	int i, j;
	HDC hdc;

	if(!ctx->editing) return;
	ctx->dirty=1;

	hdc=GetDC(ctx->hwndGame);
	x=game.field[ctx->editpos.x][ctx->editpos.y];
	switch(id) {
	case IDM_E_X: game.field[ctx->editpos.x][ctx->editpos.y]=
				(x & (I_TILE | I_ALLH | I_X   ) )^I_X;     break;
	case IDM_E_R: game.field[ctx->editpos.x][ctx->editpos.y]=
				(x & (I_TILE | I_ALLH | I_RING) )^I_RING;  break;
	case IDM_E_D: game.field[ctx->editpos.x][ctx->editpos.y]=
				(x & (I_TILE | I_ALLH | I_DOOR) )^I_DOOR;  break;
	case IDM_E_S:  // change start position
		i=game.pos.x; j=game.pos.y;
		game.field[game.pos.x][game.pos.y]=0;
		game.pos.x=ctx->editpos.x; game.pos.y=ctx->editpos.y;
		ShowLoc(ctx,hdc,i,j,0);
		break;
	case IDM_E_E: game.field[ctx->editpos.x][ctx->editpos.y]= 0; break; // erase loc.
	case IDM_E__:  // toggle tile
		if(x & I_TILE)
			game.field[ctx->editpos.x][ctx->editpos.y]= x & ~(I_TILE | I_ALLH); // turn off tile
		else
			game.field[ctx->editpos.x][ctx->editpos.y]= x | I_TILE;     // turn on tile
		break;
	case IDM_E_I: if(x&I_TILE) game.field[ctx->editpos.x][ctx->editpos.y]= x^I_HN; break;
	case IDM_E_K: if(x&I_TILE) game.field[ctx->editpos.x][ctx->editpos.y]= x^I_HE; break;
	case IDM_E_M: if(x&I_TILE) game.field[ctx->editpos.x][ctx->editpos.y]= x^I_HS; break;
	case IDM_E_J: if(x&I_TILE) game.field[ctx->editpos.x][ctx->editpos.y]= x^I_HW; break;
	case IDM_E_H: if(x&I_TILE) game.field[ctx->editpos.x][ctx->editpos.y]= x^I_ALLH; break;

	}   // end of switch()
	game.field[game.pos.x][game.pos.y]=0;
	ShowLoc(ctx,hdc,ctx->editpos.x,ctx->editpos.y,0);  // showloc will invert the tile
	ReleaseDC(ctx->hwndGame,hdc);
}

// Edit mode on/off
static void OnEdit(struct Ctx *ctx)
{
	if(WarnChanged(ctx)) return;
	ctx->editing=!ctx->editing;
	if(ctx->editing) {
		DestroyMenu(GetMenu(ctx->hwndMain));
		SetMenu(ctx->hwndMain,LoadMenu(ctx->hInst,_T("EditMenu")));
		SetWindowText(ctx->hwndMain,_T("Flop [editing]"));
		fl_NewGame(ctx);
		ctx->editpos.x=ctx->editpos.y=1;
		UpdateStat(ctx,1);
		Retile(ctx);
	}
	else {
		DestroyMenu(GetMenu(ctx->hwndMain));
		SetMenu(ctx->hwndMain,LoadMenu(ctx->hInst,_T("PlainMenu")));
		SetWindowText(ctx->hwndMain,_T("Flop"));
		LoadAndStartLevel(ctx,ctx->level);
	}
}

// Returns nonzero if command was handled.
static int OnMainCOMMAND(struct Ctx *ctx, HWND hWnd, WORD id)
{
	int x;

	switch(id) {
	case IDM_EXIT:
		if(WarnChanged(ctx)) break;
		DestroyWindow(ctx->hwndMain);
		break;
	case IDM_ABOUT:
		AboutBox(ctx,hWnd);
		break;
	case IDM_CONTENTS:
		WinHelp(ctx->hwndMain,_T("flop.hlp"),HELP_INDEX,0);
		ctx->helpused=1;
		break;

	case IDM_RESETTHIS:
		OnRESETTHIS(ctx);
		break;

	case IDM_EXPORT:
		OnEXPORT(ctx);
		break;

	case IDM_IMPORT:
		if(WarnChanged(ctx)) break;
		Import(ctx,&newgame);
		break;

	case IDM_NEXT:
		if(WarnChanged(ctx)) break;
		x=(ctx->level<100)?ctx->level+1:1;
		LoadAndStartLevel(ctx,x);
		break;

	case IDM_PREV:
		if(WarnChanged(ctx)) break;
		x=(ctx->level>1)?ctx->level-1:100;
		LoadAndStartLevel(ctx,x);
		break;

	case IDM_RELOAD:
		if(WarnChanged(ctx)) break;
		LoadAndStartLevel(ctx,ctx->level);
		break;

	case IDM_GOTOLEVEL:
		OnGOTOLEVEL(ctx);
		break;

	case IDM_SAVELEVEL:
		OnSAVELEVEL(ctx);
		break;

	case IDM_LOADMOVES:
		OnLOADMOVES(ctx);
		break;

	case IDM_SAVEMOVES:
		if(!ctx->editing) SaveMoves(ctx,ctx->move,game.moveptr);
		break;

	case IDM_SAVEMOVESFROM:
		OnSAVEMOVESFROM(ctx);
		break;

	case IDM_UNDO:
		if(!ctx->editing) Undo(ctx,game.moveptr-1);
		break;
	case IDM_UNDO10:
		if(!ctx->editing) Undo(ctx,game.moveptr-10);
		break;
	case IDM_RESTART:
		if(!ctx->editing) Undo(ctx,0);
		break;
	case IDM_REDO:
		if(!ctx->editing) Redo(ctx,1);
		break;
	case IDM_REDO10:
		if(!ctx->editing) Redo(ctx,10);
		break;
	case IDM_PLAYALL:
		if(!ctx->editing) Redo(ctx,game.movesvalid-game.moveptr);
		break;

	case IDM_GOTOMOVE:
		OnGOTOMOVE(ctx);
		break;

	case IDM_FASTREPLAY:
		OnFASTREPLAY(ctx);
		break;

	case IDM_EDITNAME:
		OnEDITNAME(ctx);
		break;

	case IDM_EDITMESSAGE:
		OnEDITMESSAGE(ctx);
		break;

	case IDM_EDITRINGS:
		OnEDITRINGS(ctx);
		break;


	case IDM_EDITCLEAR:
		OnEDITCLEAR(ctx);
		break;

	case IDM_EDITRANDOM:
		OnEDITRANDOM(ctx);
		break;

	case IDM_EDITUP:   case IDM_EDITRIGHT:   // shift field by
	case IDM_EDITDOWN: case IDM_EDITLEFT:    // one tile
		OnEditShift(ctx,(int)id-IDM_EDITUP);
		break;

	case IDM_E_X: case IDM_E_R: case IDM_E_D: case IDM_E_S:
	case IDM_E__: case IDM_E_I: case IDM_E_K: case IDM_E_M:
	case IDM_E_J: case IDM_E_H: case IDM_E_E:
		if(ctx->editing) {
			OnEditTile(ctx,id);
		}
		else {
			if(id==IDM_E_E) { // backspace key (hack)
				Undo(ctx,game.moveptr-1);
			}
		}
		break;

	case IDM_EDIT:
		OnEdit(ctx);
		break;

	default:
		return 0;
	}
	return 1; // handled
}

static void OnMainGETMINMAXINFO(struct Ctx *ctx, MINMAXINFO *mmi)
{
	// set the minimum window size
	mmi->ptMinTrackSize.x = 73+2*GetSystemMetrics(SM_CXFRAME);
	//mmi->ptMinTrackSize.y = 80;
}

static void PositionGameWindow(struct Ctx *ctx, HWND hMain)
{
	RECT rcMain;
	int x, y, cx, cy;

	if(!hMain || !ctx->hwndGame) return;

	GetClientRect(hMain,&rcMain);

	// Find the largest width such that all the tiles can be exactly the same size
	cx = ((rcMain.right-3)/TX)*TX+3;
	cy = ((rcMain.bottom-STATHEIGHT-3)/TY)*TY+3;

	// Center the game window in its parent
	x= (rcMain.right-cx)/2;
	y= (rcMain.bottom-STATHEIGHT-cy)/2;

	SetWindowPos(ctx->hwndGame,NULL,
		x,y,cx,cy,SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER);
}

static void PositionStatusWindows(struct Ctx *ctx, HWND hMain)
{
	RECT rc;

	if(!hMain) return;

	if(ctx->movewindowtextlen==0 && ctx->hwndStatus) {
		// It shouldn't really matter which window we pass to this function.
		CalcMovesWindowWidth(ctx,ctx->hwndStatus);
	}

	GetClientRect(hMain,&rc);

	if(ctx->hwndStatus) {
		SetWindowPos(ctx->hwndStatus,NULL,
			0,rc.bottom-STATHEIGHT,rc.right,STATHEIGHT,
			SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER);
	}

	if(ctx->hwndTitle) {
		SetWindowPos(ctx->hwndTitle,NULL,
			0,1,rc.right-ctx->movewindowtextlen-3,STATHEIGHT-1,
			SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER);
	}

	if(ctx->hwndMoves) {
		SetWindowPos(ctx->hwndMoves,NULL,
			rc.right-ctx->movewindowtextlen-1,1,ctx->movewindowtextlen+1,STATHEIGHT-1,
			SWP_NOACTIVATE|SWP_NOOWNERZORDER|SWP_NOZORDER);
	}
}

static void OnMainCREATE(struct Ctx *ctx, HWND hWnd)
{
	ctx->hBluePen=    CreatePen(PS_SOLID,1,RGB(0,0,128));
	ctx->hLGrayPen=	  CreatePen(PS_SOLID,1,RGB(192,192,192));
	ctx->hMGrayPen=   CreatePen(PS_SOLID,1,RGB(128,128,128));
	ctx->hTilePen=    CreatePen(PS_SOLID,1,RGB(0,255,0));
	ctx->hHandlePen=  CreatePen(PS_SOLID,1,RGB(0,0,0));   // use stock
	ctx->hRingPen=    CreatePen(PS_SOLID,1,RGB(255,255,0));
	ctx->hXPen=       CreatePen(PS_SOLID,2,RGB(0,255,255));
	ctx->hDoorPen=    CreatePen(PS_SOLID,1,RGB(255,255,255)); // stock
	ctx->hClearBrush= CreateSolidBrush(RGB(24,24,24));
	ctx->hTileBrush=  CreateSolidBrush(RGB(0,128,0));
	ctx->hDoorBrush=  CreateSolidBrush(RGB(192,128,128));
	ctx->hPlayerBrush=CreateSolidBrush(RGB(255,0,0));

	ctx->hwndStatus=CreateWindow(_T("FlopStatusClass"),_T(""),WS_CHILD|WS_CLIPCHILDREN,
		0,0,100,STATHEIGHT,hWnd,NULL,ctx->hInst,(void*)ctx);
	ctx->hwndTitle=CreateWindow(_T("FlopStatusTextClass"),_T(""),WS_CHILD,
		0,0,100,STATHEIGHT,ctx->hwndStatus,NULL,ctx->hInst,(void*)ctx);
	ctx->hwndMoves=CreateWindow(_T("FlopStatusTextClass"),_T(""),WS_CHILD,
		0,0,100,STATHEIGHT,ctx->hwndStatus,NULL,ctx->hInst,(void*)ctx);

	ctx->hwndGame=CreateWindow(_T("FlopGameClass"),_T(""),WS_CHILD,
		0,0,100,100,hWnd,NULL,ctx->hInst,(void*)ctx);

	PositionStatusWindows(ctx,hWnd);
	PositionGameWindow(ctx,hWnd);
	ShowWindow(ctx->hwndTitle,SW_SHOWNORMAL);
	ShowWindow(ctx->hwndMoves,SW_SHOWNORMAL);
	ShowWindow(ctx->hwndStatus,SW_SHOWNORMAL);
	ShowWindow(ctx->hwndGame,SW_SHOWNORMAL);
}

static void OnMainDESTROY(struct Ctx *ctx)
{
	DeleteObject(ctx->hBluePen);
	DeleteObject(ctx->hLGrayPen);
	DeleteObject(ctx->hMGrayPen);
	DeleteObject(ctx->hTilePen);
	DeleteObject(ctx->hHandlePen);
	DeleteObject(ctx->hRingPen);
	DeleteObject(ctx->hXPen);
	DeleteObject(ctx->hDoorPen);
	DeleteObject(ctx->hClearBrush);
	DeleteObject(ctx->hTileBrush);
	DeleteObject(ctx->hDoorBrush);
	DeleteObject(ctx->hPlayerBrush);
	if(ctx->helpused) WinHelp(ctx->hwndMain,_T("flop.hlp"),HELP_QUIT,0);
}

static void OnMainMOVE(struct Ctx *ctx, HWND hWnd)
{
	POINT pt;
	HDC hDC;
	HBRUSH hOldBrush;

	hDC=GetDC(hWnd);

	// TODO: These are solid brushes, and palette displays are nonexistent,
	// so I don't think we need to worry about brush origins.
	hOldBrush=SelectObject(hDC,GetStockObject(NULL_BRUSH));
	UnrealizeObject(ctx->hClearBrush);
	UnrealizeObject(ctx->hDoorBrush);
	GetDCOrgEx(hDC,&pt);
	SelectObject(hDC,ctx->hClearBrush);
	SetBrushOrgEx(hDC,pt.x%8,pt.y%8,NULL);
	SelectObject(hDC,ctx->hDoorBrush);
	SetBrushOrgEx(hDC,pt.x%8,pt.y%8,NULL);
	SelectObject(hDC,hOldBrush);
	ReleaseDC(hWnd,hDC);
}


static void OnMainSIZE(struct Ctx *ctx, HWND hWnd, WPARAM wParam)
{
	if(wParam==SIZE_MAXIMIZED || wParam==SIZE_RESTORED) {
		PositionStatusWindows(ctx,hWnd);
		PositionGameWindow(ctx,hWnd);
		CalcTileSizes(ctx);
	}
}

static void OnINITMENU(struct Ctx *ctx)
{
	int i;
	int x;
	HMENU hMenu=GetMenu(ctx->hwndMain);

	if(ctx->editing) {
		int pl=(ctx->editpos.x==game.pos.x)&&(ctx->editpos.y==game.pos.y);
		x=game.field[ctx->editpos.x][ctx->editpos.y];

		for(i=IDM_E_X;i<=IDM_E_X+4;i++)   // X,ring,door,start,tile
			EnableMenuItem(hMenu,i,pl?
				MF_GRAYED:MF_ENABLED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_E_E,((x==0)||pl)?   // erase
				MF_GRAYED:MF_ENABLED|MF_BYCOMMAND);
		for(i=IDM_E_H;i<=IDM_E_H+4;i++)     // handles
			EnableMenuItem(hMenu,i,(x&I_TILE)?
					MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
	}
	else {
		EnableMenuItem(hMenu,IDM_UNDO,game.moveptr>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_UNDO10,game.moveptr>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_RESTART,game.moveptr>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_REDO,game.moveptr<game.movesvalid?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_REDO10,game.moveptr<game.movesvalid?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_PLAYALL,game.moveptr<game.movesvalid?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_GOTOMOVE,game.movesvalid>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_LOADMOVES,(!game.levelcomplete)?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_SAVEMOVES,game.moveptr>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_SAVEMOVESFROM,game.moveptr>0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
		EnableMenuItem(hMenu,IDM_RESETTHIS,game.bestscore>=0?
				MF_ENABLED:MF_GRAYED|MF_BYCOMMAND);
	}
}

// Returns 1 if key was handled.
static int OnKEYDOWN(struct Ctx *ctx, WPARAM wParam)
{
	unsigned char arrow;
	int sx,sy;
	HDC hdc;

	if(ctx->editing) {
		sx=ctx->editpos.x; sy=ctx->editpos.y;
		switch(wParam) {
		case VK_UP:    ctx->editpos.y--; break;
		case VK_RIGHT: ctx->editpos.x++; break;
		case VK_DOWN: case VK_CLEAR:      // VK_CLEAR is numpad5
			ctx->editpos.y++; break;
		case VK_LEFT:  ctx->editpos.x--; break;
		default:
			return 0;
		}
		ctx->editpos.x = (LONG)fl_ConstrainInt(ctx->editpos.x,1,TX);
		ctx->editpos.y = (LONG)fl_ConstrainInt(ctx->editpos.y,1,TY);

		hdc=GetDC(ctx->hwndGame);
		ShowLoc(ctx,hdc,sx,sy,0);
		ShowLoc(ctx,hdc,ctx->editpos.x,ctx->editpos.y,0);
		ReleaseDC(ctx->hwndGame,hdc);
	}
	else {
		arrow=255;
		switch(wParam) {
		case VK_UP:    arrow=0; break;
		case VK_RIGHT: arrow=1; break;
		case VK_DOWN:  case VK_CLEAR:  arrow=2; break;
		case VK_LEFT:  arrow=3; break;
		default:
			return 0;
		}
		if(arrow!=255) Play(ctx,&game,&arrow,1,1);
	}
	return 1;
}

static void PaintTitleMsg(struct Ctx *ctx, HDC hDC, const RECT *rc1)
{
	RECT rc;
	rc = *rc1;

	SelectObject(hDC,ctx->hMGrayPen);
	SelectObject(hDC,ctx->hClearBrush);
	SelectObject(hDC,ctx->hStatFont);
	SetBkMode(hDC,TRANSPARENT);

	if(game.bestscore>=0) {
		StringCchPrintf(ctx->textbuf, 256, _T(" Level %d: %s   Best: %d by %s "),
			ctx->level,game.levelname,game.bestscore,game.bestname);
	}
	else {
		StringCchPrintf(ctx->textbuf, 256, _T(" Level %d: %s   Best: (none) "),
			ctx->level,game.levelname);
	}
	DrawText(hDC,ctx->textbuf,-1,&rc,DT_SINGLELINE|DT_NOPREFIX);
	//SelectObject(hDC,GetStockObject(BLACK_PEN));
}

static void PaintMovesMsg(struct Ctx *ctx, HDC hDC, const RECT *rc1)
{
	RECT rc;
	rc = *rc1;

	SelectObject(hDC,ctx->hMGrayPen);
	SelectObject(hDC,ctx->hClearBrush);
	SelectObject(hDC,ctx->hStatFont);
	SetBkMode(hDC,TRANSPARENT);

	StringCchPrintf(ctx->textbuf, 256, _T(" Rings: %d/%d   Moves: %d (%d) "),
		game.rings,game.needed,game.moves,game.moveptr);
	DrawText(hDC,ctx->textbuf,-1,&rc,DT_SINGLELINE|DT_RIGHT);
}

static void OnMainPAINT(struct Ctx *ctx, HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;

	hDC= BeginPaint(hWnd,&ps);
	EndPaint(hWnd,&ps);
}

static void OnGamePAINT(struct Ctx *ctx, HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;
	int xmax,ymax;
	RECT rect;
	int i, j;
	int xoffs=FL_XOFFS;
	int yoffs=FL_YOFFS;

	hDC= BeginPaint(hWnd,&ps);

	xmax= ctx->Lx*TX+1;
	ymax= ctx->Ly*TY+1;

	GetClientRect(hWnd,&rect);
	rect.bottom= ymax+2;          // clip off the status line

	SelectObject(hDC,ctx->hBluePen);   // blue

	for(i=0;i<TY;i++) {       // draw grid
		MoveToEx(hDC,xoffs+0   ,yoffs+i*ctx->Ly+2,NULL);
		LineTo(hDC,xoffs+xmax,yoffs+i*ctx->Ly+2);
	}
	for(i=0;i<TX;i++) {
		MoveToEx(hDC,xoffs+i*ctx->Lx+2,yoffs+0,NULL);
		LineTo(hDC,xoffs+i*ctx->Lx+2,yoffs+ymax);
	}
	SelectObject(hDC,ctx->hMGrayPen);  // gray
	for(i=0;i<=TY;i++) { // horizontal lines, bottom of grid lines
		MoveToEx(hDC,xoffs+0    -1,yoffs+i*ctx->Ly+1, NULL);
		LineTo  (hDC,xoffs+xmax +1,yoffs+i*ctx->Ly+1);
	}
	for(i=0;i<=TX;i++) { // vertical lines, right of grid lines
		MoveToEx(hDC,xoffs+i*ctx->Lx+1, yoffs+0,   NULL);
		LineTo(  hDC,xoffs+i*ctx->Lx+1, yoffs+ymax);
	}
	for(i=0;i<=TY;i++) { // horizontal lines, top of grid lines
		MoveToEx(hDC,xoffs+0    -1,yoffs+i*ctx->Ly-1, NULL);
		LineTo(  hDC,xoffs+xmax +1,yoffs+i*ctx->Ly-1);
	}
	for(i=0;i<=TX;i++) { // vertical lines, left of grid lines
		MoveToEx(hDC,xoffs+i*ctx->Lx-1, yoffs+0,   NULL);
		LineTo(  hDC,xoffs+i*ctx->Lx-1, yoffs+ymax);
	}

	SelectObject(hDC,ctx->hLGrayPen);  // light gray
	for(i=0;i<=TY;i++) {
		MoveToEx(hDC,xoffs+0   ,yoffs+i*ctx->Ly,NULL);
		LineTo(hDC,xoffs+xmax,yoffs+i*ctx->Ly);
	}
	for(i=0;i<=TX;i++) {
		MoveToEx(hDC,xoffs+i*ctx->Lx,yoffs+0,NULL);
		LineTo(hDC,xoffs+i*ctx->Lx,yoffs+ymax);
	}

	for(i=1;i<=TX;i++)       // repaint all tiles
		for(j=1;j<=TY;j++)
			ShowLoc(ctx,hDC,i,j,1);   // the 1 forces a paint

	EndPaint(hWnd,&ps);
}

static LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	struct Ctx *ctx = NULL;
	int handled=1;

	if(msg==WM_CREATE) {
		ctx = ((CREATESTRUCT*)lParam)->lpCreateParams;
		SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)ctx);
	}
	else {
		ctx = (struct Ctx*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	}

	if(!ctx) {
		handled=0;
		goto done;
	}

	switch (msg) {
	case WM_COMMAND:
		handled = OnMainCOMMAND(ctx,hWnd,LOWORD(wParam));
		break;

	case WM_GETMINMAXINFO:
		OnMainGETMINMAXINFO(ctx, (MINMAXINFO*)lParam);
		break;

	case WM_CREATE:
		OnMainCREATE(ctx,hWnd);
		break;

	case WM_DESTROY:
		OnMainDESTROY(ctx);
		PostQuitMessage(0);
		break;

	case WM_MOVE:
		OnMainMOVE(ctx,hWnd);
		break;

	case WM_INITMENU:
		OnINITMENU(ctx);
		break;

	case WM_SIZE:
		OnMainSIZE(ctx,hWnd,wParam);
		break;

	case WM_PAINT:
		OnMainPAINT(ctx,hWnd);
		break;

	case WM_KEYDOWN:
		handled = OnKEYDOWN(ctx,wParam);
		break;

	case WM_CLOSE:
		if(WarnChanged(ctx)) break;
		DestroyWindow(ctx->hwndMain);
		break;

	default:
		handled=0;
	}

done:
	if(!handled) {
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	return 0;
}


static LRESULT CALLBACK GameWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	struct Ctx *ctx = NULL;

	if(msg==WM_CREATE) {
		ctx = ((CREATESTRUCT*)lParam)->lpCreateParams;
		SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)ctx);
	}
	else {
		ctx = (struct Ctx*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	}

	if(!ctx) {
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	switch (msg) {
	case WM_LBUTTONDOWN:
		OnButtonClick(ctx,hWnd,FL_GET_X_LPARAM(lParam),FL_GET_Y_LPARAM(lParam));
		return 0;
	case WM_PAINT:
		OnGamePAINT(ctx,hWnd);
		return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}


// Returns 1 if message handled.
static int OnStatusTextPAINT(struct Ctx *ctx, HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;
	RECT rc;

	// Figure out which window we are.
	if(hWnd==ctx->hwndTitle) {
		hDC=BeginPaint(hWnd,&ps);
		GetClientRect(hWnd,&rc);
		PaintTitleMsg(ctx,hDC,&rc);
		EndPaint(hWnd,&ps);
		return 1;
	}
	else if(hWnd==ctx->hwndMoves) {
		hDC=BeginPaint(hWnd,&ps);
		GetClientRect(hWnd,&rc);
		PaintMovesMsg(ctx,hDC,&rc);
		EndPaint(hWnd,&ps);
		return 1;
	}
	return 0;
}

static void OnStatusPAINT(struct Ctx *ctx, HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;
	RECT rc;

	hDC=BeginPaint(hWnd,&ps);

	if(ctx->movewindowtextlen==0) {
		CalcMovesWindowWidth(ctx,ctx->hwndStatus);
	}

	GetClientRect(hWnd,&rc);

	SelectObject(hDC,ctx->hMGrayPen);
	MoveToEx(hDC,rc.right-ctx->movewindowtextlen-1-2,1,NULL);  // draw separator
	LineTo(hDC,rc.right-ctx->movewindowtextlen-1-2,STATHEIGHT);

	EndPaint(hWnd,&ps);
}

static LRESULT CALLBACK StatusTextWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	struct Ctx *ctx = NULL;

	if(msg==WM_CREATE) {
		ctx = ((CREATESTRUCT*)lParam)->lpCreateParams;
		SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)ctx);
	}
	else {
		ctx = (struct Ctx*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	}

	if(!ctx) {
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	switch(msg) {
	case WM_PAINT:
		OnStatusTextPAINT(ctx,hWnd);
		return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

static LRESULT CALLBACK StatusWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	struct Ctx *ctx = NULL;

	if(msg==WM_CREATE) {
		ctx = ((CREATESTRUCT*)lParam)->lpCreateParams;
		SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)ctx);
	}
	else {
		ctx = (struct Ctx*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	}

	if(!ctx) {
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}

	switch(msg) {
	case WM_PAINT:
		OnStatusPAINT(ctx,hWnd);
		return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

/****************************************************************************/

static void OnAboutINITDIALOG(struct Ctx *ctx, HWND hDlg)
{
	SetDlgItemText(hDlg, IDC_ABOUTTEXT,
		_T("Flop\r\n\r\n")
		_T("Version ") FL_VERSION_STRING _T("\r\n")
		_T("Copyright ") FL_COPYRIGHT_SYM _T(" 1995,2010 by Jason Summers\r\n\r\n")
		_T("This program is distributed under the terms of the GNU General ")
		_T("Public License, version 3 or higher.\r\n")
	  );
}

static INT_PTR CALLBACK DlgProcAbout(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	struct Ctx *ctx=NULL;

	switch (message) {
	case WM_INITDIALOG:
		ctx = (struct Ctx*)lParam;
		OnAboutINITDIALOG(ctx, hDlg);
		return TRUE;

	case WM_COMMAND:
		id=LOWORD(wParam);
		if (id==IDOK || id==IDCANCEL) {
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

/****************************************************************************/

static void init_ctx(struct Ctx *ctx)
{
	ctx->editall=0;
	ctx->editing=0;
	ctx->dirty=0;
	ctx->helpused=0;
	ctx->fastreplay=1;
	ctx->random_tile_percent=50;
	ctx->random_handle_percent=50;
}

static void fl_CreateFonts(struct Ctx *ctx)
{
	NONCLIENTMETRICS ncm;

	// Use the user's "message font", if possible.
	ZeroMemory(&ncm,sizeof(NONCLIENTMETRICS));
	ncm.cbSize = sizeof(NONCLIENTMETRICS);
	if(!SystemParametersInfo(SPI_GETNONCLIENTMETRICS,sizeof(NONCLIENTMETRICS),&ncm,0)) {
		return;
	}

	ctx->hStatFont=CreateFont(STATHEIGHT-1,0,0,0,
		FW_NORMAL,FALSE,FALSE,FALSE,ANSI_CHARSET,
		OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,DEFAULT_QUALITY,
		VARIABLE_PITCH|FF_SWISS,ncm.lfMessageFont.lfFaceName);
}

static int fl_RegisterClasses(struct Ctx *ctx)
{
	WNDCLASS  wc;
	ZeroMemory(&wc,sizeof(wc));
	wc.style = 0;
	wc.lpfnWndProc = MainWndProc;
	wc.hInstance = ctx->hInst;
	wc.hIcon = ctx->iconFlop;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName =  _T("PlainMenu");
	wc.lpszClassName = _T("FlopMainClass");
	if(0==RegisterClass(&wc)) return 0;

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = GameWndProc;
	wc.hInstance = ctx->hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _T("FlopGameClass");
	if(0==RegisterClass(&wc)) return 0;

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = StatusWndProc;
	wc.hInstance = ctx->hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _T("FlopStatusClass");
	if(0==RegisterClass(&wc)) return 0;

	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = StatusTextWndProc;
	wc.hInstance = ctx->hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(LTGRAY_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _T("FlopStatusTextClass");
	if(0==RegisterClass(&wc)) return 0;

	return 1;
}

static int fl_CreateMainWindow(struct Ctx *ctx)
{
	RECT workarea;
	int windowwidth, windowheight;

	SystemParametersInfo(SPI_GETWORKAREA,0,(void*)&workarea,0);
	windowwidth=workarea.right-workarea.left;
	windowheight=workarea.bottom-workarea.top;
	if(windowheight>windowwidth)
		windowheight=windowwidth;
	else
		windowwidth=windowheight;

	windowwidth-= GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU) + STATHEIGHT;

	ctx->hwndMain = CreateWindow(
		_T("FlopMainClass"),	/* See RegisterClass() call */
		_T("Flop"),	/* Text for window title bar */
		WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN,	/* Window style */
		workarea.left+(workarea.right-workarea.left-windowwidth)/2, /* horizontal position */
		workarea.top,   /* vertical position */
		windowwidth,    /* width */
		windowheight,   /* height */
		NULL,	/* Overlapped windows have no parent */
		NULL,	/* Use the window class menu */
		ctx->hInst,	/* This instance owns this window */
		(void*)ctx
	);

	if (!ctx->hwndMain) return 0;
	return 1;
}

static BOOL fl_InitInstance(struct Ctx *ctx, HANDLE hInstance, int nCmdShow)
{
	INITCOMMONCONTROLSEX icc;

	ctx->hInst = hInstance;

	srand((unsigned int)time(NULL));

	ZeroMemory(&icc,sizeof(INITCOMMONCONTROLSEX));
	icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
	icc.dwICC = ICC_UPDOWN_CLASS;
	InitCommonControlsEx(&icc);

	ctx->iconFlop = LoadIcon(hInstance,_T("FlopIcon"));

	if(!fl_RegisterClasses(ctx)) return FALSE;

	ctx->hAccTable = LoadAccelerators(ctx->hInst, _T("FlopAcc"));

	fl_CreateFonts(ctx);

	if(!fl_CreateMainWindow(ctx)) return FALSE;


	StringCchCopy(ctx->playername,FL_MAXNAMELEN+1,_T("Anonymous"));
	ctx->textbuf[0]='\0';     /////////
	CalcTileSizes(ctx);

	//    for(i=1;i<=TX;i++)         // maybe not necessary
	//       for(j=1;j<=TY;j++)
	//            vis[i][j]= -1;    // invalidates the screen

	ctx->level=1;
	LoadLevel(ctx,&newgame,ctx->level);
	fl_NewGame(ctx);

	/* Make the window visible; update its client area; and return "success" */

	ShowWindow(ctx->hwndMain, nCmdShow);		/* Show the window */
	UpdateWindow(ctx->hwndMain); 	/* Sends WM_PAINT message */

	return TRUE;
}

static void fl_UninitInstance(struct Ctx *ctx)
{
	if(ctx->hStatFont) DeleteObject(ctx->hStatFont);
}

int WINAPI _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	struct Ctx *ctx=NULL;

	ctx = calloc(sizeof(struct Ctx),1);
	if(!ctx) return 0;
	init_ctx(ctx);

	if (!fl_InitInstance(ctx, hInstance, nCmdShow))
		return 0;

	/* Acquire and dispatch messages until a WM_QUIT message is received. */
	while(GetMessage(&msg,NULL,0,0)) {
		if (!TranslateAccelerator(ctx->hwndMain, ctx->hAccTable, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	fl_UninitInstance(ctx);
	if(ctx) free(ctx);
	return (msg.wParam);		/* Returns the value from PostQuitMessage */
}

