// flop.h
//
// Flop is Copyright (c) 1995,2010 by Jason Summers
// This program is distributed under the terms of the GNU General
// Public License, version 3 or higher.

#define FL_VERSION_STRING _T("2.00 (beta)")

#define MOVEMAX		9999
#define STATHEIGHT  18  // status bar height, including white line at the top of it

// Structure of a saved level (bytes):
//    2 lowest score
//   21 player's name
//   21 level name
//   65 message
//  144 field
//    2 start position
// +  1 rings needed
// ----
//  256

#define TX			12  // # of horiz
#define TY			12
#define LEVELSIZE   256
#define FL_MAXNAMELEN    20 // Max length of player and level names
#define FL_MAXMESSAGELEN 64  
						
#define I_INVERTED	0x400
#define I_OPENDOOR	0x200
#define I_PLAYER	0x100

#define I_TILE  	0x080
#define I_HW		0x040
#define I_HS 		0x020
#define I_HE 		0x010
#define I_HN 		0x008
#define I_ALLH  	0x078
#define I_DOOR		0x004
#define I_X			0x002
#define I_RING		0x001

#ifdef UNICODE
#define FL_COPYRIGHT_SYM _T("\xa9")
#else
#define FL_COPYRIGHT_SYM "(c)"
#endif

struct Ctx {
	HANDLE hInst;   /* current instance */
	HACCEL hAccTable;
	HWND   hwndMain;

	HWND hwndGame;
	HWND hwndStatus; // Entire status bar, parent of hwndTitle and hwndMoves
	HWND hwndTitle; // Left part of status bar
	HWND hwndMoves; // Right part of status bar

	int editall; // Allowed to edit the first 20 levels?
	int editing;
	int level;
	int dirty;
	int helpused;

	int fastreplay;
	int Lx,Ly;     // current size of one square in pixels
	int x_handle_offs, y_handle_offs;
	int movewindowtextlen;

	POINT editpos;

	char *filebuf;
	DWORD filebuf_size;
	int filebuf_pos;

	int random_tile_percent, random_handle_percent;

	HPEN hBluePen, hLGrayPen, hMGrayPen, hTilePen;
	HPEN hHandlePen, hRingPen, hXPen, hDoorPen;
	HBRUSH hTileBrush, hClearBrush, hDoorBrush, hPlayerBrush;
	HFONT hStatFont;
	HICON iconFlop;

	TCHAR textbuf[256];
	TCHAR buffer[100];
	TCHAR playername[FL_MAXNAMELEN+1];
	int vis[TX+2][TY+2];   // stores what is actually on the screen
	unsigned char move[MOVEMAX];  // movelist
	unsigned char move2[MOVEMAX]; // temp movelist
};

struct Game {
	int field[TX+2][TY+2];
	POINT pos;  /* player location */
	int moveptr;     // number of moves made, counting moving tiles
	int movesvalid;
	int moves;       // number of moves made
	int rings;       // number of rings collected so far
	int xs;    // number of X's that are not covered
	int tileselected;
	POINT seltile;
	int levelcomplete;

	int needed;      // number of rings needed to exit
	TCHAR levelname[FL_MAXNAMELEN+1];
	TCHAR message[FL_MAXMESSAGELEN+1];
	int bestscore;
	TCHAR bestname[FL_MAXNAMELEN+1];
};

// flop.c
void Error(struct Ctx *ctx, const TCHAR *m);
void fl_InitGame(struct Game *g);
void UpdateAll(struct Ctx *ctx);
void fl_NewGame(struct Ctx *ctx);

// flop-io.c
int LoadLevel(struct Ctx *ctx, struct Game *g, int levelnum);
int SaveLevel(struct Ctx *ctx, struct Game *g,int b,int scoreonly);
int Import(struct Ctx *ctx, struct Game *newgame);
int Export(struct Ctx *ctx, struct Game *newgame);
int LoadMoves(struct Ctx *ctx, unsigned char *m, int m_size);
int SaveMoves(struct Ctx *ctx, unsigned char *m, int n);
